local addon, shared = ...

local RELEASE=		99
local TESTING=		20
local WDOGYIELD=	15
local FUNCTIONRUN=	10
local STATUSUPDT=	8
local INFORMATION=	5
local DETAIL=		1

local RAT_INACTIVE=	1
local RAT_ACTIVE=	2

---------------------
-- Variables
---------------------

_Rattus = {}

local RX = _Rattus

local UIF = shared.UIFUNC

local _achvupdates = {}
local _achvtoggle = {}
local _srchdata = {}
local _srchOffset = 0
local _scrollwin = 2

local _watchdogyield = {}

local unscaled = {
	h16 = 0,
	h14 = 0,
	h12 = 0,
	hh = 0, hw = 0, tw = 0
	}

local scaled = {}
	
RX.UIElements = {
	header = {},
	tracker = {
		prev = nil,
		ACHV = {},
		cnt = 0, active = 0,
		elements = {}, inuse = {}
	},
	search = { achv={} },
	controlbar = {}, setlist = { set = {} }
}

RX.achv = {
	all = {},
	avl = {},
	map = {}
}

RX.achvcat = {
}
RX.sortachvcat = {
}

RX.scScroll = 0
RX.mcScroll = 0
local maincatID={}
local subcatID={}
local maincatSelected=nil
local subcatSelected=nil

local default_settings = {
	x = 500, y = 500, v=1.91, activeSet = "", selectedSet = "",
	saved = { },
	hidecomp = false, showcomp = false,
	progtext = true, progbar = true, autozone = false,
	showdelete = true, scale = 1, hidden = false, mmoptions = true
}

local svLoaded = false
local coroDead = false
local searchVis = false
local controlVis = false
local setlistVis = false

local rtx_UIBuild = true
local rtx_AchvList = false
local rtx_AchvListDone = false
local rtx_LoadSet = false
local rtx_LoadSetName = ""
local rtx_ToggleComp = false
local rtx_AchvUpdate = false
local rtx_SearchUpdt = false
local rtx_SetChanged = false
local rtx_ZoneChange = false
local rtx_ZoneChangeName = ""

local _playerID = 0

---------------------
-- Functions
---------------------

local debuglevel = RELEASE

local function Debug(s,l)
	if (l or INFORMATION) >= debuglevel then
		print(s)
	end
end

local yieldcnt = 0
local function WATCHDOGWAIT(x)
	local rmx = Inspect.System.Watchdog()
	while rmx < 0.03 do
		yieldcnt = yieldcnt+1
		Debug("#"..yieldcnt..":rmx: "..rmx.." ["..x.."]",WDOGYIELD)
		_watchdogyield[yieldcnt] = { tx = os.date("%Y%m%d-%H:%M:%S"), m = x, tr = rmx }
		coroutine.yield()
		rmx = Inspect.System.Watchdog()
	end
end

local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if o[k] == nil then
				o[k] = {}
			end
	 	 	if type(o[k]) == 'table' then
	 			MergeTable(o[k], n[k])
	 	 	end
		else
			if o[k] == nil then
				o[k] = v
			end
		end
	end
end

local function _GetTimeStamp()
	local _tr = Inspect.Time.Real()
	_tr = (_tr - math.floor(_tr)) * 1000000
	return string.format("%s.%d", os.date("%Y%m%d%H%M%S"), _tr)
end

function RX.SetChanged()
	local actvSet
	
	if RX.UIElements.search.window:GetVisible() then
		actvSet = RX.UIElements.search.dropdown:GetText()
	else
		actvSet = Rattus_Settings.activeSet
	end
	for id,v in pairs(_achvtoggle) do
		if Rattus_Settings.saved[actvSet][id] == nil then
			Rattus_Settings.saved[actvSet][id] = _GetTimeStamp()
			if actvSet == Rattus_Settings.activeSet then
				RX.AddTracker(id)
			end
		else
			Rattus_Settings.saved[actvSet][id] = nil
			if actvSet == Rattus_Settings.activeSet then
				RX.RemoveTracker(id)
			end
		end
		_achvtoggle[id] = nil
	end
end

function RX.ToggleAchv(x)
	local id = _srchdata[x+_srchOffset].v
	Debug(string.format("#%d clicked. ID = %s", x, _srchdata[x].v), DETAIL)
	_achvtoggle[id] = true
	rtx_SetChanged = true
end

function RX.controlBar_MouseOut(self,h)
	local md = Inspect.Mouse()
	if setlistVis then
		local t = RX.UIElements.setlist.frame:GetTop()
		local b = RX.UIElements.setlist.frame:GetBottom()
		local l = RX.UIElements.setlist.frame:GetLeft()
		local r = RX.UIElements.setlist.frame:GetRight()
		local mmt = RX.UIElements.header.fldr:GetTop()
		local mmb = RX.UIElements.header.fldr:GetBottom()
		local mml = RX.UIElements.header.fldr:GetLeft()
		local mmr = RX.UIElements.header.fldr:GetRight()
		if (md.x < l or md.x > r or md.y < t or md.y > b) and (md.x < mml or md.x > mmr or md.y < mmt or md.y > b) then
			RX.UIElements.setlist.frame:SetVisible(false)
			setlistVis = false
		end
	elseif controlVis then
		local t = RX.UIElements.controlbar.frame:GetTop()
		local b = RX.UIElements.controlbar.frame:GetBottom()
		local l = RX.UIElements.controlbar.frame:GetLeft()
		local r = RX.UIElements.controlbar.frame:GetRight()

		if Rattus_Settings.mmoptions then
			local mmt = RX.UIElements.minimap:GetTop()
			local mmb = RX.UIElements.minimap:GetBottom()
			local mml = RX.UIElements.minimap:GetLeft()
			local mmr = RX.UIElements.minimap:GetRight()

			if (md.x < l or md.x > r or md.y < t or md.y > b) and (md.x < mml or md.x > mmr or md.y < mmt or md.y > mmb) then
				RX.UIElements.controlbar.frame:SetVisible(false)
				controlVis = false
			end
		else
			if md.x < l or md.x > r or md.y < t or md.y > b then
				RX.UIElements.controlbar.frame:SetVisible(false)
				controlVis = false
			end
		end
	end
end

function RX.BuildUIElements()
	RX.context = UI.CreateContext(addon.identifier)

	WATCHDOGWAIT("BuildUIElements:0")

	RX.UIElements.minimap = UIF.mk_texture("RX.UIElements.minimap",RX.context,36,36)
	RX.UIElements.minimap:SetTexture(addon.identifier, "textures/mm_button.png")

	if MINIMAPDOCKER then
		MINIMAPDOCKER.Register(addon.identifier, RX.UIElements.minimap)
	else
		RX.UIElements.minimap:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
			self.MouseDown = true
			local mouseData = Inspect.Mouse()
			self.sx = mouseData.x - RX.UIElements.minimap:GetLeft()
			self.sy = mouseData.y - RX.UIElements.minimap:GetTop()
		end, "Event.UI.Input.Mouse.Left.Down")

		RX.UIElements.minimap:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
			if self.MouseDown then
				local nx, ny
				local mouseData = Inspect.Mouse()
				nx = mouseData.x - self.sx
				ny = mouseData.y - self.sy
				RX.UIElements.minimap:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
			end
		end, "Event.UI.Input.Mouse.Cursor.Move")

		RX.UIElements.minimap:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
			if self.MouseDown then
				self.MouseDown = false
			end
			Rattus_Settings.mmx = RX.UIElements.minimap:GetLeft()
			Rattus_Settings.mmy = RX.UIElements.minimap:GetTop()
		end, "Event.UI.Input.Mouse.Left.Up")
	end

	RX.UIElements.minimap:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		Rattus_Settings.hidden = not Rattus_Settings.hidden
		if Rattus_Settings.hidden then
			RX.UIElements.header.icon:SetVisible(false)
			RX.UIElements.search.window:SetVisible(false)
			RX.UIElements.search.filter:SetKeyFocus(false)
		else
			RX.UIElements.search.window:SetVisible(searchVis)
			RX.UIElements.header.icon:SetVisible(true)
		end
	end, "Event.UI.Input.Mouse.Right.Click")

	WATCHDOGWAIT("f:1")

	RX.UIElements.header.icon = UIF.mk_texture("RX.UIElements.header.icon",RX.context,nil,nil,1)
	RX.UIElements.header.icon:SetVisible(false)
	RX.UIElements.header.icon:SetBackgroundColor(0,0,0,0.5)
	RX.UIElements.header.icon:SetTexture("Rift", "medal_gold.png.dds")
	RX.UIElements.header.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", default_settings.x, default_settings.y)
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.mouseDown = true
		local md = Inspect.Mouse()
		self.sx = md.x - RX.UIElements.header.icon:GetLeft()
		self.sy = md.y - RX.UIElements.header.icon:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")	
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.mouseDown then
			local nx, ny
			local md = Inspect.Mouse()
			nx = md.x - self.sx
			ny = md.y - self.sy
			RX.UIElements.header.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx, ny)
			Rattus_Settings.x = nx
			Rattus_Settings.y = ny
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.mouseDown then
			self.mouseDown = false
		end
	end, "Event.UI.Input.Mouse.Left.Up")	
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		searchVis = not searchVis
		RX.UIElements.search.window:SetVisible(searchVis)
		rtx_SearchUpdt = true
	end, "Event.UI.Input.Mouse.Right.Click")
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		Rattus_Settings.scale = Rattus_Settings.scale + 0.05
		RX.ScaleUI()
	end, "Event.UI.Input.Mouse.Wheel.Forward")
	RX.UIElements.header.icon:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		Rattus_Settings.scale = Rattus_Settings.scale - 0.05
		RX.ScaleUI()
	end, "Event.UI.Input.Mouse.Wheel.Back")

	WATCHDOGWAIT("BuildUIElements:1a")

	RX.UIElements.controlbar.frame = UIF.mk_frame("RX.UIElements.controlbar.frame",RX.context,nil,nil,10,nil,false)
	RX.UIElements.controlbar.frame:SetBackgroundColor(0,0,0.5,1)
	RX.UIElements.controlbar.frame:SetMouseMasking("limited")
	RX.UIElements.controlbar.frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")

	WATCHDOGWAIT("BuildUIElements:2")

	RX.UIElements.header.zoom = UIF.mk_texture("RX.UIElements.header.zoom",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.zoom:SetVisible(true)
	RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom.png")
	RX.UIElements.header.zoom:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		local dx = Rattus_Settings.hidecomp
		dx = not dx
		Debug("Hiding completed objectives: "..tostring(dx), RELEASE)
		Rattus_Settings.hidecomp = dx
		if dx == true then
			RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom_bw.png")
		else
			RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom.png")
		end
		rtx_ToggleComp = true
	end, "Event.UI.Input.Mouse.Left.Click")
	RX.UIElements.header.zoom:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")
	WATCHDOGWAIT("BuildUIElements:2a")

	RX.UIElements.header.text = UIF.mk_texture("RX.UIElements.header.text",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go.png")
	RX.UIElements.header.text:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		Rattus_Settings.progtext = not Rattus_Settings.progtext
		Debug(string.format("Showing progress as text: %s", tostring(Rattus_Settings.progtext)), RELEASE)
		if Rattus_Settings.progtext == true then
			RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go.png")
		else
			RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go_bw.png")
		end
		rtx_ToggleComp = true
	end, "Event.UI.Input.Mouse.Left.Click")
	RX.UIElements.header.text:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")

	WATCHDOGWAIT("BuildUIElements:2b")

	RX.UIElements.header.bars = UIF.mk_texture("RX.UIElements.header.bars",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar.png")
	
	RX.UIElements.header.bars:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		Rattus_Settings.progbar = not Rattus_Settings.progbar
		Debug(string.format("Showing progress bar: %s", tostring(Rattus_Settings.progbar)), RELEASE)
		if Rattus_Settings.progbar == true then
			RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar.png")
		else
			RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar_bw.png")
		end

		rtx_ToggleComp = true
	end, "Event.UI.Input.Mouse.Left.Click")

	RX.UIElements.header.bars:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")
	WATCHDOGWAIT("BuildUIElements:2c")

	RX.UIElements.header.map = UIF.mk_texture("RX.UIElements.header.map",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go.png")
	RX.UIElements.header.map:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		Rattus_Settings.autozone = not Rattus_Settings.autozone
		Debug(string.format("Autoloading set=ZONENAME: %s", tostring(Rattus_Settings.autozone)), RELEASE)
		if Rattus_Settings.autozone == true then
			RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go.png")
			RX.HandleZoneChange(LIBZONECHANGE.currentZoneName)
		else
			RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go_bw.png")
			if Rattus_Settings.selectedSet ~= Rattus_Settings.activeSet then
				Debug(string.format("LC: %s != %s, loading %s", Rattus_Settings.activeSet, Rattus_Settings.selectedSet, Rattus_Settings.selectedSet), INFORMATION)
				rtx_LoadSetName = Rattus_Settings.selectedSet
				rtx_LoadSet = true
			end
		end
	end, "Event.UI.Input.Mouse.Left.Click")
	RX.UIElements.header.map:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		RX.Command_Slash_Register(0, "new ZONE")
	end, "Event.UI.Input.Mouse.Right.Click")

	RX.UIElements.header.map:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")
	WATCHDOGWAIT("BuildUIElements:2c")

	RX.UIElements.header.fldr = UIF.mk_texture("RX.UIElements.header.fldr",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.fldr:SetTexture(addon.identifier, "textures/folder_explore.png")
	RX.UIElements.header.fldr:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
		_sets = {}
		for k,v in pairs(Rattus_Settings.saved) do
			table.insert(_sets, k)
		end
		table.sort(_sets)
		local prev = nil

		for k,v in ipairs(_sets) do
			if RX.UIElements.setlist.set[v] == nil then
				local f = {}
				f.frame = UIF.mk_frame("setFrame",RX.UIElements.setlist.frame,unscaled.hw,unscaled.h14,32)
				f.frame:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
					rtx_LoadSetName = v
					rtx_LoadSet = true
				end, "Event.UI.Input.Mouse.Left.Click")
				f.frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
					f.frame:SetBackgroundColor(0,1,0)
				end, "Event.UI.Input.Mouse.Cursor.In")
				f.frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
					f.frame:SetBackgroundColor(0,0,0)
					RX.controlBar_MouseOut(0,0)
				end, "Event.UI.Input.Mouse.Cursor.Out")

				f.text = UIF.mk_text("setText",RX.UIElements.setlist.frame,unscaled.f14,33,v)
				f.text:SetHeight(unscaled.h14)
				f.text:SetPoint("CENTER", f.frame, "CENTER")
				RX.UIElements.setlist.set[v] = f
			end
			if prev == nil then
				RX.UIElements.setlist.set[v].frame:SetPoint("TOPLEFT", RX.UIElements.setlist.frame, "TOPLEFT", 0,0)
			else
				RX.UIElements.setlist.set[v].frame:SetPoint("TOPLEFT", prev, "BOTTOMLEFT", 0,0)
			end
			prev = RX.UIElements.setlist.set[v].frame
		end
		RX.UIElements.setlist.frame:SetHeight(#_sets * unscaled.h14)
		setlistVis = true
		RX.UIElements.setlist.frame:SetVisible(true)
	end, "Event.UI.Input.Mouse.Cursor.In")
	RX.UIElements.header.fldr:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")
	WATCHDOGWAIT("BuildUIElements:2d")

	RX.UIElements.header.frame = UIF.mk_frame("RX.UIElements.header.frame",RX.UIElements.header.icon,nil,nil,-5)
	RX.UIElements.header.frame:SetPoint("TOPLEFT", RX.UIElements.header.icon, "TOPRIGHT", 0,0)
	RX.UIElements.header.frame:SetBackgroundColor(0,0,0,0.5)
	RX.UIElements.header.frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
		if Rattus_Settings.mmoptions == false then
			RX.UIElements.controlbar.frame:SetVisible(true)
			controlVis = true
		end
	end, "Event.UI.Input.Mouse.Cursor.In")

	RX.UIElements.minimap:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
		if Rattus_Settings.mmoptions then
			RX.UIElements.controlbar.frame:SetVisible(true)
			controlVis = true
		end
	end, "Event.UI.Input.Mouse.Cursor.In")

	WATCHDOGWAIT("BuildUIElements:3")

	RX.UIElements.header.title = UIF.mk_text("RX.UIElements.header.title",RX.UIElements.header.icon,16,1,"RATTUS")
	unscaled.f16 = 16
	unscaled.h16 = RX.UIElements.header.title:GetHeight()
	RX.UIElements.header.title:SetPoint("TOPCENTER", RX.UIElements.header.frame, "TOPCENTER", 0,0)
	WATCHDOGWAIT("BuildUIElements:4")

	RX.UIElements.header.active = UIF.mk_text("RX.UIElements.header.active",RX.UIElements.header.icon,14,1,"Set:")
	unscaled.h14 = RX.UIElements.header.active:GetHeight()
	unscaled.f14 = 14
	RX.UIElements.header.active:SetFontSize(12)
	unscaled.h12 = RX.UIElements.header.active:GetHeight()
	unscaled.f12 = 12
	RX.UIElements.header.active:SetPoint("BOTTOMCENTER", RX.UIElements.header.frame, "BOTTOMCENTER", 0,0)
	WATCHDOGWAIT("BuildUIElements:5")

	RX.UIElements.header.ffor = UIF.mk_texture("RX.UIElements.header.ffor",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.ffor:SetTexture(addon.identifier, "textures/control_fastforward.png")
	RX.UIElements.header.ffor:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RX.FindSet("F")
	end, "Event.UI.Input.Mouse.Left.Click")
	RX.UIElements.header.ffor:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")

	RX.UIElements.header.dele = UIF.mk_texture("RX.UIElements.header.dele",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.dele:SetTexture(addon.identifier, "textures/bin_empty.png")
	RX.UIElements.header.dele:EventAttach(Event.UI.Input.Mouse.Right.Click, function(self, h)
		RX.Delete()
	end, "Event.UI.Input.Mouse.Right.Click")
	RX.UIElements.header.dele:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")

	WATCHDOGWAIT("BuildUIElements:6")

	RX.UIElements.header.rewd = UIF.mk_texture("RX.UIElements.header.rewd",RX.UIElements.controlbar.frame,nil,nil,15)
	RX.UIElements.header.rewd:SetTexture(addon.identifier, "textures/control_rewind.png")
	RX.UIElements.header.rewd:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RX.FindSet("P")
	end, "Event.UI.Input.Mouse.Left.Click")
	RX.UIElements.header.rewd:EventAttach(Event.UI.Input.Mouse.Cursor.Out, RX.controlBar_MouseOut, "Event.UI.Input.Mouse.Cursor.Out")

	WATCHDOGWAIT("BuildUIElements:7")
	
	unscaled.hh = unscaled.h12 + unscaled.h16
	local fx70 = math.floor(unscaled.hh * 0.70)

	RX.UIElements.header.rewd:SetPoint("LEFTCENTER", RX.UIElements.controlbar.frame, "LEFTCENTER", 0,0)
	RX.UIElements.header.zoom:SetPoint("TOPLEFT", RX.UIElements.header.rewd, "TOPRIGHT", 0,0)
	RX.UIElements.header.text:SetPoint("TOPLEFT", RX.UIElements.header.zoom, "TOPRIGHT", 0,0)
	RX.UIElements.header.bars:SetPoint("TOPLEFT", RX.UIElements.header.text, "TOPRIGHT", 0,0)
	RX.UIElements.header.map:SetPoint("TOPLEFT", RX.UIElements.header.bars, "TOPRIGHT", 0,0)
	RX.UIElements.header.fldr:SetPoint("TOPLEFT", RX.UIElements.header.map, "TOPRIGHT", 0,0)
	RX.UIElements.header.dele:SetPoint("TOPRIGHT", RX.UIElements.header.ffor, "TOPLEFT", 0,0)
	RX.UIElements.header.ffor:SetPoint("RIGHTCENTER", RX.UIElements.controlbar.frame, "RIGHTCENTER", 0,0)

	WATCHDOGWAIT("BuildUIElements:7a")

	RX.UIElements.setlist.frame = UIF.mk_frame("RX.UIElements.setlist.frame",RX.UIElements.header.icon,nil,nil,30,a,false)
	RX.UIElements.setlist.frame:SetBackgroundColor(0,0,0,1)
	RX.UIElements.setlist.frame:SetPoint("TOPLEFT", RX.UIElements.controlbar.frame, "BOTTOMLEFT", 0,0)

	WATCHDOGWAIT("BuildUIElements:7b")

	RX.UIElements.header.dele:SetHeight(fx70)
	RX.UIElements.header.dele:SetWidth(fx70)
	RX.UIElements.header.ffor:SetHeight(fx70)
	RX.UIElements.header.ffor:SetWidth(fx70)
	RX.UIElements.header.rewd:SetHeight(fx70)
	RX.UIElements.header.rewd:SetWidth(fx70)
	RX.UIElements.header.icon:SetHeight(unscaled.hh)
	RX.UIElements.header.icon:SetWidth(unscaled.hh)
	RX.UIElements.header.zoom:SetHeight(fx70)
	RX.UIElements.header.zoom:SetWidth(fx70)
	RX.UIElements.header.text:SetHeight(fx70)
	RX.UIElements.header.text:SetWidth(fx70)
	RX.UIElements.header.bars:SetHeight(fx70)
	RX.UIElements.header.bars:SetWidth(fx70)
	RX.UIElements.header.map:SetHeight(fx70)
	RX.UIElements.header.map:SetWidth(fx70)
	RX.UIElements.header.fldr:SetHeight(fx70)
	RX.UIElements.header.fldr:SetWidth(fx70)

	WATCHDOGWAIT("BuildUIElements:7c")

	unscaled.tw = RX.UIElements.header.title:GetWidth()
	RX.UIElements.header.frame:SetHeight(unscaled.hh)
	unscaled.hw = unscaled.tw * 4
	RX.UIElements.header.frame:SetWidth(unscaled.hw)

	RX.UIElements.controlbar.frame:SetHeight(unscaled.hh)
	RX.UIElements.controlbar.frame:SetWidth(unscaled.hw)

	RX.UIElements.setlist.frame:SetWidth(unscaled.hw)

	RX.UIElements.header.icon:SetVisible(true)
	WATCHDOGWAIT("BuildUIElements:8")

	local sw = 900

	RX.UIElements.search.window = UIF.mk_frame("RX.UIElements.search.window",RX.context,sw,nil,1,nil,searchVis)
	RX.UIElements.search.window:SetBackgroundColor(0.2, 0.2, 0.2, 0.75)
	RX.UIElements.search.window:SetPoint("CENTER", UIParent, "CENTER", 0,0)

	RX.UIElements.search.window:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self, h)
		self.mouseDown = true
		local md = Inspect.Mouse()
		self.sx = md.x - RX.UIElements.search.window:GetLeft()
		self.sy = md.y - RX.UIElements.search.window:GetTop()
	end, "Event.UI.Input.Mouse.Left.Down")	
	RX.UIElements.search.window:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function(self, h)
		if self.mouseDown then
			local nx, ny
			local md = Inspect.Mouse()
			nx = md.x - self.sx
			ny = md.y - self.sy
			RX.UIElements.search.window:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx, ny)
			Rattus_Settings.srchx = nx
			Rattus_Settings.srchy = ny
		end
	end, "Event.UI.Input.Mouse.Cursor.Move")
	RX.UIElements.search.window:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self, h)
		if self.mouseDown then
			self.mouseDown = false
		end
	end, "Event.UI.Input.Mouse.Left.Up")
	RX.UIElements.search.window:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function(self, h)
		if _scrollwin==1 then
			if RX.scScroll > 0 then
				RX.scScroll = RX.scScroll -1
			end
		elseif _scrollwin==2 then
			if _srchOffset > 0 then
				_srchOffset = _srchOffset -1
			end
		end
		rtx_SearchUpdt = true
	end, "Event.UI.Input.Mouse.Wheel.Forward")
	RX.UIElements.search.window:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function(self, h)
		if _scrollwin==1 and RX.scScroll + (15+1) < RX.scScrollMax then
			RX.scScroll = RX.scScroll +1
		elseif _scrollwin==2 then
			_srchOffset = _srchOffset +1
		end
		rtx_SearchUpdt = true
	end, "Event.UI.Input.Mouse.Wheel.Back")

	WATCHDOGWAIT("BuildUIElements:8a")

	RX.UIElements.search.close = UIF.mk_texture("RX.UIElements.search.close",RX.UIElements.search.window,unscaled.h16,unscaled.h16,5)
	RX.UIElements.search.close:SetTexture(addon.identifier, "textures/cancel.png")
	RX.UIElements.search.close:SetPoint("TOPRIGHT", RX.UIElements.search.window, "TOPRIGHT", 0,0)
	RX.UIElements.search.close:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
		RX.UIElements.search.filter:SetKeyFocus(false)
		searchVis = false
		RX.UIElements.search.window:SetVisible(searchVis)
	end, "Event.UI.Input.Mouse.Left.Click")

	WATCHDOGWAIT("BuildUIElements:8b")

	RX.UIElements.search.header = UIF.mk_text("RX.UIElements.search.header",RX.UIElements.search.window,16,3,"Rattus Search...")
	RX.UIElements.search.header:SetBackgroundColor(0,0,0.5)
	RX.UIElements.search.header:SetPoint("TOPLEFT", RX.UIElements.search.window, "TOPLEFT", 0,0)
	RX.UIElements.search.header:SetWidth(sw)
	RX.UIElements.search.mc={}
	RX.UIElements.search.sc={}
	RX.UIElements.search.achv={}
	
	RX.UIElements.search.dropdown = UIF.mk_dropdown("RX.UIElements.search.dropdown",RX.UIElements.search.window,50,200,"Set:",{},nil, false)
	RX.UIElements.search.dropdown:SetPoint("TOPCENTER", RX.UIElements.search.window, "TOPCENTER")
	
	for x=1, 15 do
		for f=0, 2 do
			if f==0 then
				a=RX.UIElements.search.mc
			elseif f==1 then
				a=RX.UIElements.search.sc
			else
				a=RX.UIElements.search.achv
			end
			WATCHDOGWAIT("BuildUIElements:9:"..x..","..f)
			local fx = UIF.mk_text("RX.UIElements.search.achv:"..x..","..f,RX.UIElements.search.window,14,10,"LINE: "..x..","..f)
			fx:SetWidth(sw/3-20)
			local bg = x/15
			fx:SetBackgroundColor(0,0,0,0.25)
			if x==1 then
				fx:SetPoint("TOPLEFT",  RX.UIElements.search.header, "BOTTOMLEFT", f*sw/3+10,5)
			else
				fx:SetPoint("TOPLEFT",  a[x-1], "BOTTOMLEFT", 0,0)
			end
			if (f==0) then
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
					fx:SetBackgroundColor(0,0.5,0,1)
				end, "Event.UI.Input.Mouse.Cursor.In")
				fx:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
					if (maincatSelected == maincatID[x]) then
						maincatSelected=nil
					else
						maincatSelected=maincatID[x]
					end
					RX.scScroll=0
					subcatSelected=nil
					rtx_SearchUpdt=true
				end, "Event.UI.Input.Mouse.Left.Click")
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
					if (maincatSelected == maincatID[x]) then
						fx:SetBackgroundColor(0,0,0.5,0.8)
					else
						fx:SetBackgroundColor(0,0,0,0.25)
					end
				end, "Event.UI.Input.Mouse.Cursor.Out")
			elseif (f==1) then
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
					fx:SetBackgroundColor(0,0.5,0,1)
					_scrollwin=1
				end, "Event.UI.Input.Mouse.Cursor.In")
				fx:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
					if (subcatSelected == subcatID[x]) then
						subcatSelected=nil
					else
						subcatSelected=subcatID[x]
					end
					rtx_SearchUpdt=true
				end, "Event.UI.Input.Mouse.Left.Click")
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
					if (subcatSelected == subcatID[x]) then
						fx:SetBackgroundColor(0,0,0.5,0.8)
					else
						fx:SetBackgroundColor(0,0,0,0.25)
					end
				end, "Event.UI.Input.Mouse.Cursor.Out")
			elseif (f==2) then
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(self, h)
					local id = _srchdata[x+_srchOffset].v
					RX.UIElements.search.descript:SetText(RX.achv.all[id].description)
					fx:SetBackgroundColor(0,0.5,0,1)
					_scrollwin=2
				end, "Event.UI.Input.Mouse.Cursor.In")
				fx:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
					local id = _srchdata[x+_srchOffset].v
					local actvSet = RX.UIElements.search.dropdown:GetText()
					local achvComp = _srchdata[x+_srchOffset].complete
					local achvInc = Rattus_Settings.saved[actvSet][id]
					
					RX.ToggleAchv(x)

					if achvComp == true then
						if achvInc then
							RX.UIElements.search.achv[x]:SetFontColor(0.5, 1, 0.5)
						else
							RX.UIElements.search.achv[x]:SetFontColor(0.25, 0.5, 0.25)
						end
					else
						if achvInc then
							RX.UIElements.search.achv[x]:SetFontColor(1, 1, 1)
						else
							RX.UIElements.search.achv[x]:SetFontColor(0.5, 0.5, 0.5)
						end
					end
				end, "Event.UI.Input.Mouse.Left.Click")
				fx:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(self, h)
					fx:SetBackgroundColor(0,0,0,0.25)			
				end, "Event.UI.Input.Mouse.Cursor.Out")
			end
			a[x] = fx
		end
	end

	WATCHDOGWAIT("BuildUIElements:10")

	RX.UIElements.search.label = UIF.mk_text("RX.UIElements.search.label",RX.UIElements.search.window,12,5,"Filter:")
	RX.UIElements.search.label:SetPoint("TOPLEFT", RX.UIElements.search.mc[15], "BOTTOMLEFT", 0,10)

	RX.UIElements.search.filter = UI.CreateFrame("RiftTextfield", "RX.UIElements.search.filter", RX.UIElements.search.window)
	RX.UIElements.search.filter:SetWidth((sw/2)+20)
	RX.UIElements.search.filter:SetText("")
	RX.UIElements.search.filter:SetLayer(10)
	RX.UIElements.search.filter:SetPoint("TOPLEFT", RX.UIElements.search.label, "TOPRIGHT", 5,0)
	RX.UIElements.search.filter:SetBackgroundColor(0.3, 0.1, 0.3)

--	RX.UIElements.search.filter:EventAttach(Event.UI.Textfield.Change, function(self, h)
--		_srchOffset = 0
--		rtx_SearchUpdt = true
--	end, "Event.UI.Textfield.Change")
	
	RX.UIElements.search.filter:EventAttach(Event.UI.Input.Key.Up, function(self, h)
		_srchOffset = 0
		rtx_SearchUpdt = true
	end, "Event.UI.Input.Key.Up")
	
	WATCHDOGWAIT("BuildUIElements:11")

	RX.UIElements.search.chkboxlbl = UIF.mk_text("RX.UIElements.search.chkboxlbl",RX.UIElements.search.window,nil,5,"Show Completed:")
	RX.UIElements.search.chkboxlbl:SetPoint("TOPLEFT", RX.UIElements.search.filter, "TOPRIGHT", 10,0)

	RX.UIElements.search.chkbox = UI.CreateFrame("RiftCheckbox", "RX.UIElements.search.chkbox", RX.UIElements.search.window)
	RX.UIElements.search.chkbox:SetLayer(5)
	RX.UIElements.search.chkbox:SetHeight(20)
	RX.UIElements.search.chkbox:SetWidth(20)
	RX.UIElements.search.chkbox:SetPoint("TOPLEFT", RX.UIElements.search.chkboxlbl, "TOPRIGHT", 5,0)
	RX.UIElements.search.chkbox:EventAttach(Event.UI.Checkbox.Change, function(self, h)
		Rattus_Settings.showcomp = RX.UIElements.search.chkbox:GetChecked()
		rtx_SearchUpdt = true
	end, "Event.UI.Checkbox.Change")

	RX.UIElements.search.descript = UIF.mk_text("RX.UIElements.search.descript",RX.UIElements.search.window,12,5,"")
	RX.UIElements.search.descript:SetBackgroundColor(0,0,0,0.25)
	RX.UIElements.search.descript:SetPoint("TOPLEFT", RX.UIElements.search.label, "BOTTOMLEFT", 0,10)

	RX.UIElements.search.window:SetHeight(unscaled.h16 + (15 * unscaled.h14) + (2 * unscaled.h12) + 35)
end

function RX.BuildAchvList()
	Debug("Entering BuildAchvList. Need player.", STATUSUPDT)
	local needPlayer = true
	local nx = 0
	while needPlayer do
		nx = nx+1
		RX.UIElements.header.title:SetText("RATTUS: Need player details: "..nx)
		local d = Inspect.Unit.Detail("player")
		if d ~= nil then
			if d.alliance ~= nil then
				RX.alliance = d.alliance
				needPlayer = false
			else
				if d.race == "highelf" or d.race == "mathosian" or d.race == "dwarf" then
					RX.alliance = "guardian"
					needPlayer = false
				elseif d.race == "bahmi" or d.race == "kelari" or d.race == "eth" then
					RX.alliance = "defiant"
					needPlayer = false
				else
					coroutine.yield()
				end
			end
		else
			coroutine.yield()
		end
	end
	Debug("Got player. Finding achvt details", STATUSUPDT)
	RX.UIElements.header.title:SetText("RATTUS: Need achv details")
	local ix = 0
	local cnt=0
	nx=0
	local achvlist = Inspect.Achievement.List()
	while achvlist == nil do
		nx = nx+1
		RX.UIElements.header.title:SetText("RATTUS: Need achv details. #"..nx)
		coroutine.yield()
		achvlist = Inspect.Achievement.List()
	end
	
	for k,v in pairs(achvlist) do
		cnt=cnt+1
	end
	
	for k,v in pairs(Inspect.Achievement.Category.Detail(Inspect.Achievement.Category.List())) do
		RX.achvcat[k] = v
		table.insert(RX.sortachvcat, { id=k, name=v.name })
	end
	
	RX.UIElements.header.title:SetText("RATTUS: Details: 0/"..cnt)
	for k,v in pairs(achvlist) do
		
		ix = ix+1
		if (ix%10 == 0) then
			RX.UIElements.header.title:SetText("RATTUS: Details: "..ix.."/"..cnt)
		end
		WATCHDOGWAIT("BuildAchvList:1:"..ix)
		local d = Inspect.Achievement.Detail(k)
		while d == nil do
			print(string.format("nil details for %s", k))
			coroutine.yield()
			d = Inspect.Achievement.Detail(k)
		end

		if (d.alliance or RX.alliance) == RX.alliance then
			RX.achv.all[k] = d
			if d.complete ~= true then
				RX.achv.avl[k] = d
			end
			if RX.achv.map[d.name] ~= nil then
				local c = RX.achv.map[d.name].count
				c=c+1
				RX.achv.map[d.name].count = c
				RX.achv.map[d.name].details[c] = k
			else
				RX.achv.map[d.name] = { count=1, details = { [1] = k }}
			end
		end
	end
	RX.UIElements.header.title:SetText("RATTUS")
	Debug("Found all details.", STATUSUPDT)
	
	WATCHDOGWAIT("BuildAchvList:sort:")
	table.sort(RX.sortachvcat, function(a, b)
		-- print ("compare "..(a.name or "nil") .. " with " .. (b.name or "nil"))
		return (a.name or "") < (b.name or "")
	end)
end

function RX.reset()
	local x = Rattus_Settings.x
	local y = Rattus_Settings.y

	Rattus_Settings = {}

	MergeTable(Rattus_Settings, Rattus_Settings)

	Rattus_Settings.x = x
	Rattus_Settings.y = y
end

function RX.ConvertSettings(s_from)
	-- v1.9 switched from names to achv ids
	if s_from < 1.9 then
		if Rattus_Settings.saved ~= nil then
			Rattus_Settings.tmp_saved = {}
			MergeTable(Rattus_Settings.tmp_saved, Rattus_Settings.saved)
			Rattus_Settings.saved = {}
			for k,v in pairs(Rattus_Settings.tmp_saved) do
				Rattus_Settings.saved[k] = {}
				for k1, v1 in pairs(v) do
					if RX.achv.map[v1] ~= nil then
						if RX.achv.map[v1].count > 1 then
							Debug(string.format("NOTICE: '%s/%s' is ambiguous. Adding all (%d) possible matches.", k, v1, RX.achv.map[v1].count), RELEASE)
						end
						for _,dv in pairs(RX.achv.map[v1].details) do
							Rattus_Settings.saved[k][dv] = true
						end
					else
						Debug(string.format("Unable to convert: %s", v1), RELEASE)
					end
				end
			end
		end
		Rattus_Settings.showcomp = true
		Rattus_Settings.v = 1.9
	end
	-- v1.91 sort based on time added
	if s_from < 1.91 then
		if Rattus_Settings.saved ~= nil then
			for k,v in pairs(Rattus_Settings.saved) do
				for k1,v1 in pairs(Rattus_Settings.saved[k]) do
					Rattus_Settings.saved[k][k1] = nil
					Rattus_Settings.saved[k][k1] = _GetTimeStamp()
				end
			end
		end
		Rattus_Settings.v = 1.91
	end
end

function RX.FindSet(d)
	local _sets = {}
	for k,v in pairs(Rattus_Settings.saved) do
		table.insert(_sets, k)
	end
	table.sort(_sets)
	local nxt
	for x=1,#_sets do
		if _sets[x] == Rattus_Settings.activeSet then
			nxt = x
			break
		end
	end
	if d == "F" then
		nxt = nxt+1
		if nxt > #_sets then
			nxt = 1
		end
	elseif d == "P" then
		nxt = nxt-1
		if nxt < 1 then 
			nxt = #_sets
		end
	end
	rtx_LoadSetName = _sets[nxt]
	Rattus_Settings.selectedSet = rtx_LoadSetName
	rtx_LoadSet = true
end

function RX.RemoveTracker(id)
	Rattus_Settings.saved[Rattus_Settings.activeSet][id] = nil
	RX.UIElements.tracker.ACHV[id].frame:SetVisible(false)

	RX.UIElements.tracker.prev = nil

	local _set = {}
	for k,v in pairs(Rattus_Settings.saved[Rattus_Settings.activeSet]) do
		table.insert(_set, {i=k, t=v})
	end
	table.sort(_set, function(a,b) return a.t < b.t end)

	for _,v in ipairs(_set) do
		local k = v.i
		if RX.UIElements.tracker.prev == nil then
			RX.UIElements.tracker.ACHV[k].frame:SetPoint("TOPLEFT", RX.UIElements.header.icon, "BOTTOMLEFT", 0,0)
		else
			RX.UIElements.tracker.ACHV[k].frame:SetPoint("TOPLEFT", RX.UIElements.tracker.ACHV[RX.UIElements.tracker.prev].frame, "BOTTOMLEFT", 0,0)
		end
		RX.UIElements.tracker.prev = k
	end
end

function RX.AddTracker(id)
	WATCHDOGWAIT("AddTracker:1:"..id)
	local d = RX.achv.all[id]
	local template = { reqs = {} }
	local frmWidth = scaled.hw+scaled.h12+scaled.h16
	
	if RX.UIElements.tracker.ACHV[id] == nil then
		Debug(string.format("Building UI frames for %s", id), INFORMATION)
		template.frame = UIF.mk_frame("frm"..id,RX.UIElements.header.icon,frmWidth,scaled.h14,1,nil,false)
		WATCHDOGWAIT("AddTracker:2:"..id)
		template.icon = UIF.mk_texture("icn"..id,template.frame,scaled.h14,scaled.h14,5)
		template.icon:SetPoint("TOPLEFT", template.frame, "TOPLEFT", 0,0)
		template.icon:SetTexture("Rift", d.icon)
		template.icon:EventAttach(Event.UI.Input.Mouse.Left.Click, function(self, h)
			RX.RemoveTracker(id)
		end, "Event.UI.Input.Mouse.Left.Click")
		WATCHDOGWAIT("AddTracker:3:"..id)
		template.name = UIF.mk_text("nam"..id,template.frame,scaled.f14,5,d.name)
		template.name:SetPoint("TOPLEFT", template.icon, "TOPRIGHT", 0,0)
		template.name:SetFontColor(1,1,1)
		WATCHDOGWAIT("AddTracker:4:"..id)

		local _frmaxwid = math.floor(frmWidth * 0.75)

		for k,v in pairs(d.requirement) do
			WATCHDOGWAIT("AddTracker:5:"..id..":"..k)
			template.reqs[k] = {}
			template.reqs[k].name = UIF.mk_text("req"..k,template.frame,scaled.f12,5,v.name or "NO OBJECTIVE")
			if template.reqs[k].name:GetWidth() > _frmaxwid then
				template.reqs[k].trunc = true
				template.reqs[k].longname = UIF.mk_text("lreq"..k,template.frame,scaled.f12,5)
				template.reqs[k].longname:SetVisible(false)
				template.reqs[k].longname:SetPoint("TOPLEFT", template.reqs[k].name, "TOPLEFT", 0,0)
				local s = v.name or "NO OBJECTIVE"
				local l = string.len(s)
				for x=l, 1, -1 do
					local ns = string.format("%s...", string.sub(s,1,x))
					template.reqs[k].name:SetText(ns)
					if template.reqs[k].name:GetWidth() <= _frmaxwid then
						break
					end
				end
				for x=l, 1, -1 do
					local ns = string.format("%s...", string.sub(s,1,x))
					template.reqs[k].longname:SetText(ns)
					if template.reqs[k].longname:GetWidth() <= scaled.hw then
						break
					end
				end
			else
				template.reqs[k].trunc = false
			end
			template.reqs[k].name:SetFontColor(1,1,1)
			template.reqs[k].name:SetWidth(frmWidth * 0.75)

			template.reqs[k].prog = UIF.mk_text("req"..k,template.frame,scaled.f12,5,"")
			template.reqs[k].prog:SetPoint("TOPRIGHT", template.reqs[k].name, "TOPLEFT", frmWidth,0)

			template.reqs[k].bar = UIF.mk_frame("bar"..k,template.frame,nil,scaled.h12,3)
			template.reqs[k].bar:SetPoint("TOPLEFT", template.reqs[k].name, "TOPLEFT", 0,0)
			template.reqs[k].bar:SetBackgroundColor(0, 0.5, 0, 0.5)
		end

		function template.update()
			local d = Inspect.Achievement.Detail(id)
			RX.achv.all[id] = d
			WATCHDOGWAIT("update:"..id)
			local last = 0
			local cnt = 0
			if d.complete == true then
				template.name:SetFontColor(0,1,0)
			end

			local frmW = scaled.h16 + scaled.h12 + scaled.hw

			for k,v in pairs(d.requirement) do
				if v.complete == true then
					template.reqs[k].name:SetFontColor(0,1,0)
				end

				local pct = 0

				if (v.count or 0) > 0 then
					if v.complete == true then
						template.reqs[k].prog:SetText(string.format("%d/%s", v.count, v.count))
						pct = 1
					else
						template.reqs[k].prog:SetText(string.format("%d/%s", (v.countDone or 0), v.count))
						pct = (v.countDone or 0)/v.count
					end
				else
					if v.complete == true then
						pct = 1
					else
						pct = 0
					end
				end

				if (v.complete ~= true) or (Rattus_Settings.hidecomp == false) then
					cnt = cnt+1
					if last == 0 then
						template.reqs[k].name:SetPoint("TOPLEFT", template.icon, "BOTTOMLEFT", 0,0)
					else
						template.reqs[k].name:SetPoint("TOPLEFT", template.reqs[last].name, "BOTTOMLEFT", 0,0)
					end
					last = k
					if template.reqs[k].trunc == true then
						if Rattus_Settings.progtext == true then
							template.reqs[k].name:SetVisible(true)
							template.reqs[k].longname:SetVisible(false)
						else
							template.reqs[k].name:SetVisible(false)
							template.reqs[k].longname:SetVisible(true)
						end
					else
						template.reqs[k].name:SetVisible(true)
					end
					template.reqs[k].prog:SetVisible(Rattus_Settings.progtext)
					template.reqs[k].bar:SetVisible(Rattus_Settings.progbar)

				else
					template.reqs[k].name:SetVisible(false)
					if template.reqs[k].trunc == true then
						template.reqs[k].longname:SetVisible(false)
					end
					template.reqs[k].prog:SetVisible(false)
					template.reqs[k].bar:SetVisible(false)
				end

				template.reqs[k].bar:SetWidth(frmW * pct)
			end
			template.frame:SetHeight(scaled.h14 + (cnt * scaled.h12))
		end
		RX.UIElements.tracker.ACHV[id] = template
	end

	RX.UIElements.tracker.ACHV[id].update()

	if RX.UIElements.tracker.prev == nil then
		RX.UIElements.tracker.ACHV[id].frame:SetPoint("TOPLEFT", RX.UIElements.header.icon, "BOTTOMLEFT", 0,0)
	else
		RX.UIElements.tracker.ACHV[id].frame:SetPoint("TOPLEFT", RX.UIElements.tracker.ACHV[RX.UIElements.tracker.prev].frame, "BOTTOMLEFT", 0,0)
	end

	RX.UIElements.tracker.prev = id
	RX.UIElements.tracker.ACHV[id].frame:SetVisible(true)
end

function RX.ToggleComp()
	for k,v in pairs(Rattus_Settings.saved[Rattus_Settings.activeSet]) do
		RX.UIElements.tracker.ACHV[k].update()
	end
end

function RX.LoadSet(set)
	if Rattus_Settings.saved[set] == nil then
		print(string.format("ERROR: No such set: %s", set))
		return
	end

	Debug(string.format("Loading... %s", set), RELEASE)
	if Rattus_Settings.activeSet ~= nil then
		if Rattus_Settings.saved[Rattus_Settings.activeSet] ~= nil then
			for k,v in pairs(Rattus_Settings.saved[Rattus_Settings.activeSet]) do
				if RX.UIElements.tracker.ACHV[k] ~= nil then
					RX.UIElements.tracker.ACHV[k].frame:SetVisible(false)
				end
			end
		end
	end
	Rattus_Settings.activeSet = set
	RX.UIElements.tracker.prev = nil
	RX.UIElements.header.active:SetText(string.format("Set: %s", set))

	local _set = {}


	for k,v in pairs(Rattus_Settings.saved[set]) do
		table.insert(_set, {i=k, t=v})
	end
	table.sort(_set, function(a,b) return a.t < b.t end)

	for _,v in ipairs(_set) do
		local k = v.i
		if RX.achv.all[k] ~= nil then
			Debug(string.format("Adding achv: %s/%s [%s]", set, RX.achv.all[k].name, k), STATUSUPDT)
			RX.AddTracker(k)
--		else
--			Rattus_Settings.saved[set][k] = nil
--			print("removing... "..k)
		end
	end
end

function RX.updateDDList()
	local _sets = {}
	for k,v in pairs(Rattus_Settings.saved) do
		table.insert(_sets, {n=k})
	end
	table.sort(_sets, function(a,b) return a.n < b.n end)
	RX.UIElements.search.dropdown.Update(_sets)
end

function RX.ScaleUI()

	scaled = {}
	_Rattus.scaled = scaled
	for k,v in pairs(unscaled) do
		scaled[k] = math.floor(v * Rattus_Settings.scale)
	end

	RX.UIElements.header.title:SetFontSize(scaled.f16)
	scaled.h16 = RX.UIElements.header.title:GetHeight()
	
	RX.UIElements.header.active:SetFontSize(scaled.f14)
	scaled.h14 = RX.UIElements.header.active:GetHeight()
	RX.UIElements.header.active:SetFontSize(scaled.f12)
	scaled.h12 = RX.UIElements.header.active:GetHeight()

	scaled.tw = RX.UIElements.header.title:GetWidth()
	scaled.hw = scaled.tw * 4
	scaled.hh = scaled.h16 + scaled.h12
	
	RX.UIElements.header.icon:SetWidth(scaled.hh)
	RX.UIElements.header.icon:SetHeight(scaled.hh)
	
	for k,v in pairs(RX.UIElements.tracker.ACHV) do
		v.frame:SetHeight(scaled.h14)
		v.frame:SetWidth(scaled.hh+scaled.hw)
		v.icon:SetHeight(scaled.h14)
		v.icon:SetWidth(scaled.h14)
		v.name:SetFontSize(scaled.f14)
		for k1,v1 in pairs(v.reqs) do
			v1.name:SetFontSize(scaled.f12)
			if v1.longname then
				v1.longname:SetFontSize(scaled.f12)
			end
			v1.prog:SetFontSize(scaled.f12)
			v1.prog:SetPoint("TOPRIGHT", v1.name, "TOPLEFT", (scaled.hh+scaled.hw),0)
			v1.bar:SetHeight(scaled.h12)
		end
		v.update()
	end

	RX.UIElements.header.frame:SetHeight(scaled.hh)
	RX.UIElements.header.frame:SetWidth(scaled.hw)
	RX.UIElements.setlist.frame:SetWidth(scaled.hw)
end

function RX.SVLoaded()
	if Rattus_Settings ~= nil then
		if (Rattus_Settings.v or 0) < default_settings.v then
			Debug("Need to convert saved sets...", RELEASE)
			RX.ConvertSettings(Rattus_Settings.v or 0)
		end
	else
		Rattus_Settings = {}
	end
	if not MINIMAPDOCKER then
		if Rattus_Settings.mmx ~= nil then
			RX.UIElements.minimap:SetPoint("TOPLEFT", UIParent, "TOPLEFT", Rattus_Settings.mmx, Rattus_Settings.mmy)
		else
			RX.UIElements.minimap:SetPoint("TOPLEFT", UI.Native.MapMini, "BOTTOMCENTER", 0,0)
			Rattus_Settings.mmx = RX.UIElements.minimap:GetLeft()
			Rattus_Settings.mmy = RX.UIElements.minimap:GetTop()
		end
	end
	if Rattus_Settings.srchx == nil then
		Rattus_Settings.srchx = RX.UIElements.search.window:GetLeft()
		Rattus_Settings.srchy = RX.UIElements.search.window:GetTop()
	end
	RX.UIElements.search.window:ClearPoint("CENTER")
	RX.UIElements.search.window:SetPoint("TOPLEFT", UIParent, "TOPLEFT", Rattus_Settings.srchx, Rattus_Settings.srchy)

	MergeTable(Rattus_Settings, default_settings)

	RX.UIElements.header.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", Rattus_Settings.x, Rattus_Settings.y)
	RX.UIElements.search.chkbox:SetChecked(Rattus_Settings.showcomp)
	if Rattus_Settings.activeSet == "" then
		local s = os.date("%Y%m%d-%H%M%S")
		Rattus_Settings.saved[s] = {}
		Rattus_Settings.activeSet = s
	end
	if Rattus_Settings.selectedSet == "" then
		Rattus_Settings.selectedSet = Rattus_Settings.activeSet
	end
	RX.UIElements.header.dele:SetVisible(Rattus_Settings.showdelete)

	if Rattus_Settings.hidecomp == true then
		RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom_bw.png")
	else
		RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom.png")
	end
	if Rattus_Settings.progtext == true then
		RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go.png")
	else
		RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go_bw.png")
	end
	if Rattus_Settings.progbar == true then
		RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar.png")
	else
		RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar_bw.png")
	end
	if Rattus_Settings.autozone == true then
		RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go.png")
	else
		RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go_bw.png")
	end
	if Rattus_Settings.hidden then
		RX.UIElements.header.icon:SetVisible(false)
		RX.UIElements.search.window:SetVisible(false)
		RX.UIElements.search.filter:SetKeyFocus(false)
	else
		RX.UIElements.search.window:SetVisible(searchVis)
		RX.UIElements.header.icon:SetVisible(true)
	end

	if Rattus_Settings.mmoptions then
		RX.UIElements.controlbar.frame:SetPoint("TOPLEFT", RX.UIElements.minimap, "CENTER")
	else
		RX.UIElements.controlbar.frame:SetPoint("TOPLEFT", RX.UIElements.header.icon, "TOPRIGHT", 0,0)
	end

	RX.ScaleUI()
	
	RX.updateDDList()
	RX.UIElements.search.dropdown:SetText(Rattus_Settings.activeSet)
	
	if rtx_LoadSetName == "" then
		rtx_LoadSetName = Rattus_Settings.activeSet
		rtx_LoadSet = true
	end
end

function RX.Event_Achievement_Complete(h,id)
	_achvupdates[id] = true
	rtx_AchvUpdate = true
end

function RX.Event_Achievement_Update(h,ids)
	Debug("RX.AchvUpdate", FUNCTIONRUN)
	local ix=0
	for k,v in pairs(ids) do
		_achvupdates[k] = true
		ix=ix+1
	end
	Debug("#="..ix, DETAIL)
	rtx_AchvUpdate = true
end

__BADACHV = {}
__CMPACHV = {}

function RX.HandleAchvUpdate()
	local cnt = 0
	local fac = 0
	local cmp = 0
	local bad = 0
	for k,v in pairs(_achvupdates) do
		cnt = cnt+1
		WATCHDOGWAIT("HandleUpdate:1:"..cnt)
		local d = Inspect.Achievement.Detail(k)
		if d ~= nil then
			if (d.alliance or RX.alliance) == RX.alliance then
				fac = fac+1
				RX.achv.all[k] = d
				if Rattus_Settings.saved[Rattus_Settings.activeSet][k] ~= nil then
					Debug("ID: "..k.." in active list...", DETAIL)
					RX.UIElements.tracker.ACHV[k].update()
				end
				if d.complete == true then
					cmp = cmp+1
					__CMPACHV[cmp] = k
				end
			else
				bad = bad+1
				__BADACHV[bad] = k
			end
			if RX.GADGET.ACTIVEACHVT[k] then
				for ak, av in ipairs(RX.GADGET.ACTIVEACHVT[k]) do
					RX.GADGET.HandleUpdate(k, av)
				end
			end
			_achvupdates[k] = nil
		end
	end
	Debug(string.format("#=%d, fac=%d, bad=%d, cmp=%d", cnt, fac, bad, cmp), INFORMATION)
end

function RX.HandleZoneChange(z)
	if Rattus_Settings == nil or rtx_UIBuild == true then
		rtx_ZoneChangeName = z
		rtx_ZoneChange = true
	else
		rtx_ZoneChange = false
		if Rattus_Settings.autozone then
			if Rattus_Settings.saved[z] then
				Debug("found set: "..z, DETAIL)
				rtx_LoadSetName = z
				rtx_LoadSet = true
			else
				if Rattus_Settings.activeSet ~= Rattus_Settings.selectedSet then
					if Rattus_Settings.saved[Rattus_Settings.selectedSet] ~= nil then
						Debug(string.format("%s != %s, loading %s", Rattus_Settings.activeSet, Rattus_Settings.selectedSet, Rattus_Settings.selectedSet), TESTING)
						rtx_LoadSetName = Rattus_Settings.selectedSet
						rtx_LoadSet = true
					else
						local s = ""
						for k,v in pairs(Rattus_Settings.saved) do
							s = k
							break
						end
						if s == "" then
							s = os.date("%Y%m%d-%H%M%S")
							Rattus_Settings.saved[s] = {}
						end
						Debug(string.format("%s != %s. %s missing, loading %s", Rattus_Settings.activeSet, Rattus_Settings.selectedSet, Rattus_Settings.selectedSet, s), TESTING)
						Rattus_Settings.selectedSet = s
						rtx_LoadSetName = s
						rtx_LoadSet = true
					end
				end
			end
		else
			Debug("autozone disabled not loading", INFORMATION)
		end
	end
end

function RX.UpdateSearchWindow() 

	if searchVis == false then return end

	local t1 = Inspect.Time.Real()
	local comp = RX.UIElements.search.chkbox:GetChecked()
	local filt = string.lower(RX.UIElements.search.filter:GetText())

	local src
	local dat

	local cnt = 0
	
	mc=1
	sc=1
	maincatID={}
	subcatID={}
	for i, k in ipairs(RX.sortachvcat) do
		v=RX.achvcat[k.id]
		if v.parent and (maincatSelected==nil or maincatSelected==v.parent) then
			x = sc - RX.scScroll
			if x>=1 and x<=15 then
				RX.UIElements.search.sc[x]:SetText(v.name)
				RX.UIElements.search.sc[x]:SetVisible(true)
				if k.id == subcatSelected then
					RX.UIElements.search.sc[x]:SetBackgroundColor(0,0,0.5,0.8)
				else
					RX.UIElements.search.sc[x]:SetBackgroundColor(0,0,0,0.25)
				end
				subcatID[x]=k.id
			end
			sc = sc + 1
			RX.scScrollMax=sc
		elseif v.parent==nil then
			x = mc - RX.mcScroll 
			if x>=1 and x<=15 then
				RX.UIElements.search.mc[x]:SetText(v.name)
				if k.id == maincatSelected then
					RX.UIElements.search.mc[x]:SetBackgroundColor(0,0,0.5,0.8)
				else
					RX.UIElements.search.mc[x]:SetBackgroundColor(0,0,0,0.25)
				end
				maincatID[x]=k.id
			end
			mc = mc + 1
		end
	end
	for x=sc - RX.scScroll, 15 do
		RX.UIElements.search.sc[x]:SetVisible(false)
	end
	for x=mc - RX.mcScroll, 15 do
		RX.UIElements.search.mc[x]:SetVisible(false)
	end

	if comp then
		src = RX.achv.all
	else
		src = RX.achv.avl
	end

	_srchdata = {}

	for k,v in pairs(src) do
		local s = string.lower(v.name)
		local t = string.lower(v.description)
		if (((maincatSelected==nil and subcatSelected==nil)
		  or (v.category==maincatSelected and subcatSelected ==nil)
		  or (v.category==subcatSelected))
		and (filt=="" or string.find(s, filt) or string.find(t, filt))) then
			cnt = cnt + 1
			table.insert(_srchdata, { n = v.name, v = k, complete=v.complete })
		end
	end

	table.sort(_srchdata, function(a,b) return a.n < b.n end)

	local ubnd

	if #_srchdata <= 15 then
		_srchOffset = 0
		ubnd = math.min(15, #_srchdata)
	elseif #_srchdata-_srchOffset < 15 then
		_srchOffset = #_srchdata - 15
		ubnd = 15
	else
		ubnd = 15
	end
	
	local actvSet = RX.UIElements.search.dropdown:GetText()

	for x=1, ubnd do
		RX.UIElements.search.achv[x]:SetText(_srchdata[x+_srchOffset].n)
		if _srchdata[x+_srchOffset].complete == true then
			if Rattus_Settings.saved[actvSet][_srchdata[x+_srchOffset].v] then
				RX.UIElements.search.achv[x]:SetFontColor(0.25, 0.5, 0.25)
			else
				RX.UIElements.search.achv[x]:SetFontColor(0.5, 1, 0.5)
			end
		else
			if Rattus_Settings.saved[actvSet][_srchdata[x+_srchOffset].v] then
				RX.UIElements.search.achv[x]:SetFontColor(0.5, 0.5, 0.5)
			else
				RX.UIElements.search.achv[x]:SetFontColor(1, 1, 1)
			end
		end
		RX.UIElements.search.achv[x]:SetVisible(true)
	end
	for x=ubnd+1, 15 do
		RX.UIElements.search.achv[x]:SetVisible(false)
	end

	local t2 = Inspect.Time.Real()
	Debug(string.format("Search completed in %fs. #=%d", (t2-t1), cnt), DETAIL)
end

local _frcount = 0

function RX.UpdateHandler()
	while true do
		if rtx_UIBuild then
			Debug("building UI frames",WDOGYIELD)
			RX.BuildUIElements()
			rtx_UIBuild = false
		elseif rtx_AchvList then
			Debug("building achv list",WDOGYIELD)
			if rtx_AchvListDone == false then
				RX.BuildAchvList()
			end
			rtx_AchvListDone = true
			rtx_AchvList = false
		elseif svLoaded then
			Debug("SV Loaded",WDOGYIELD)
			RX.SVLoaded()
			svLoaded = false
		elseif (rtx_LoadSet and rtx_AchvListDone) then
			Debug("load set",WDOGYIELD)
			RX.LoadSet(rtx_LoadSetName)
			rtx_LoadSet = false
		elseif rtx_ToggleComp then
			RX.ToggleComp()
			rtx_ToggleComp = false
		elseif rtx_ZoneChange then
			RX.HandleZoneChange(rtx_ZoneChangeName)
		elseif rtx_AchvUpdate then
			RX.HandleAchvUpdate()
			rtx_AchvUpdate = false
		elseif rtx_SetChanged then
			RX.SetChanged()
			rtx_SetChanged = false
		elseif rtx_SearchUpdt then
			RX.UpdateSearchWindow()
			rtx_SearchUpdt = false
		elseif rtx_DeleteSet then
			Rattus_Settings.saved[rtx_DeleteName] = nil
			rtx_DeleteSet = false
			RX.updateDDList()
		elseif controlVis or setlistVis then
			_frcount = _frcount+1
			if _frcount == 10 then
				RX.controlBar_MouseOut(0,0)
				_frcount = 0
			end
		end
		coroutine.yield()
	end
end

RX.coUpdateHandler = coroutine.create(RX.UpdateHandler)

function RX.Event_System_Update_Begin(h)
	if coroDead then return end
	local ok, errormsg = coroutine.resume(RX.coUpdateHandler)
	if not ok then
		coroDead = true
		error("Error in coroutine: "..errormsg)
	end
end

function RX.Delete()
	local s = ""
	for k,v in pairs(Rattus_Settings.saved) do
		if k ~= Rattus_Settings.activeSet then
			s = k
			break
		end
	end
	if s == "" then
		s = os.date("%Y%m%d-%H%M%S")
		Rattus_Settings.saved[s] = {}
	end
	Debug(string.format("Deleting set: %s, Making %s active",Rattus_Settings.activeSet,s), RELEASE)
	rtx_LoadSetName = s
	rtx_LoadSet = true
	rtx_DeleteName = Rattus_Settings.activeSet
	RX.updateDDList()
	RX.UIElements.search.dropdown:SetText(s)
	rtx_DeleteSet = true
end

function RX.showUsage()
	Debug("/rattus hidecomplete -- toggles display of completed objectives", RELEASE)
	Debug("/rattus autozone -- toggles auto change of set based on zonename", RELEASE)
	Debug("/rattus progress text -- toggles display of progress text", RELEASE)
	Debug("/rattus progress bars -- toggles display of progress bars", RELEASE)
	Debug("/rattus add NAME -- Adds the achievement(s) called NAME to the current set", RELEASE)
	Debug("/rattus delete NAME -- Removes the achievement(s) called NAME to the current set", RELEASE)
	Debug("/rattus load NAME -- makes the set NAME active", RELEASE)
	Debug("/rattus copy NAME -- copies the current set to NAME", RELEASE)
	Debug("/rattus new [ZONE|NAME] -- creates a new set [called NAME, or the name of your current zone]", RELEASE)
	Debug("/rattus rename NAME -- renames the current set to NAME", RELEASE)
	Debug("/rattus deleteset [NAME] -- deletes the set NAME, or current set if no name given", RELEASE)
	Debug("/rattus list -- lists all currents sets", RELEASE)
end

function RX.Command_Slash_Register(h, args)
	local r = {}
	local numargs = 0
	for token in string.gmatch(args, "[^%s]+") do
		r[numargs] = token
		numargs=numargs+1
	end
	if numargs>0 then
		if r[0] == "hidecomplete" then
			local dx = Rattus_Settings.hidecomp
			dx = not dx
			Debug("Hiding completed objectives: "..tostring(dx), RELEASE)
			if dx == true then
				RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom_bw.png")
			else
				RX.UIElements.header.zoom:SetTexture(addon.identifier, "textures/zoom.png")
			end
			Rattus_Settings.hidecomp = dx
			rtx_ToggleComp = true
			return
		elseif r[0] == "new" then
			local s
			if numargs >= 2 then
				if numargs == 2 and r[1] == "ZONE" then
					s = LIBZONECHANGE.currentZoneName
				else
					s = r[1]
					for i = 2, numargs-1 do
						s = s.." "..r[i]
					end
				end
			else
				s = os.date("%Y%m%d-%H%M%S")
			end

			if Rattus_Settings.saved[s] == nil then
				Rattus_Settings.saved[s] = {}
				Debug(string.format("Creating new set: %s", s), RELEASE)
				if r[1] ~= "ZONE" then
					Rattus_Settings.selectedSet = s
				end
				rtx_LoadSetName = s
				RX.updateDDList()
				rtx_LoadSet = true
			else
				Debug(string.format("Set: %s already exists. Please delete first.",s), RELEASE)
			end
		elseif r[0] == "clear" then
			for k,v in pairs(Rattus_Settings.saved[Rattus_Settings.activeSet]) do
				RX.RemoveTracker(k)
			end
		elseif r[0] == "copy" then
			if numargs >= 2 then
				local s = r[1]
				for i = 2, numargs-1 do
					s = s.." "..r[i]
				end
				Debug(string.format("Copying %s to %s", Rattus_Settings.activeSet, s), RELEASE)
				Rattus_Settings.saved[s] = {}
				RX.updateDDList()
				MergeTable(Rattus_Settings.saved[s], Rattus_Settings.saved[Rattus_Settings.activeSet])
			else
				Debug("Need to give a target name.", RELEASE)
			end
		elseif r[0] == "deleteset" then
			if numargs >= 2 then
				local s = r[1]
				for i = 2, numargs-1 do
					s = s.." "..r[i]
				end
				if Rattus_Settings.saved[s] == nil then
					Debug(string.format("No such saved set: %s", s))
				else
					if s == Rattus_Settings.activeSet then
						RX.Delete()
					else
						Debug(string.format("Deleting set: %s",s), RELEASE)
						Rattus_Settings.saved[s] = nil
					end
				end
			else
				RX.Delete()
			end
		elseif r[0] == "rename" then
			if numargs >= 2 then
				local s = r[1]
				for i = 2, numargs-1 do
					s = s.." "..r[i]
				end
				Debug(string.format("Renaming set: %s, as %s",Rattus_Settings.activeSet,s), RELEASE)
				Rattus_Settings.saved[s] = {}
				MergeTable(Rattus_Settings.saved[s], Rattus_Settings.saved[Rattus_Settings.activeSet])
				rtx_LoadSetName = s
				rtx_LoadSet = true
				rtx_DeleteName = Rattus_Settings.activeSet
				rtx_DeleteSet = true
			else
				Debug("Need to give a target name.", RELEASE)
			end
		elseif r[0] == "autozone" then
			Rattus_Settings.autozone = not Rattus_Settings.autozone
			Debug(string.format("Autoloading set=ZONENAME: %s", tostring(Rattus_Settings.autozone)), RELEASE)
			if Rattus_Settings.autozone == true then
				RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go.png")
				RX.HandleZoneChange(LIBZONECHANGE.currentZoneName)
			else
				RX.UIElements.header.map:SetTexture(addon.identifier, "textures/map_go_bw.png")
				if Rattus_Settings.selectedSet ~= Rattus_Settings.activeSet then
					rtx_LoadSetName = Rattus_Settings.selectedSet
					rtx_LoadSet = true
				end
			end
		elseif r[0] == "list" then
			local x = 1
			Debug("Saved sets:", RELEASE)
			for k,_ in pairs(Rattus_Settings.saved) do
				Debug(string.format("%d : %s", x, k), RELEASE)
				x=x+1
			end
		elseif r[0] == "load" then
			if numargs >= 2 then
				local s = r[1]
				for i = 2, numargs-1 do
					s = s.." "..r[i]
				end
				if not Rattus_Settings.saved[s] then
					Debug("No such set currently saved!", RELEASE)
				else
					Rattus_Settings.selectedSet = s
					rtx_LoadSetName = s
					rtx_LoadSet = true
				end
			else
				Debug("Need to give an achievement set name to load", RELEASE)
			end
		elseif (r[0] == "add") or (r[0] == "delete") then
			if numargs >= 2 then
				local s = r[1]
				for i = 2, numargs-1 do
					s = s.." "..r[i]
				end
				local ls= string.lower(s)
				if string.sub(ls, 1, 1) == "[" and string.sub(ls, -1) == "]" then ls = string.sub(ls, 2, -2) end
				local added = false
				local found = false
				for k,v in pairs(RX.achv.all) do
					if string.lower(v.name) == ls then
						found = true
						if Rattus_Settings.saved[Rattus_Settings.activeSet][k] == nil then
							if r[0] == "add" then
								_achvtoggle[k] = true
								added = true
							end
						else
							if r[0] == "delete" then
								_achvtoggle[k] = true
								added = true
							end

						end
					end
				end
				if added then
					rtx_SetChanged = true
				elseif found == false then
					Debug(string.format("Unable to find achievement: %s", s), RELEASE)
				end
			else
				Debug("Need to give an achievement name to search for.", RELEASE)
			end
		elseif r[0] == "watchdog" then
				for x=1, yieldcnt do
					Debug(string.format("%d @%s [%.4f] %s", x, _watchdogyield[x].tx, _watchdogyield[x].tr, _watchdogyield[x].m), RELEASE)
				end
		elseif r[0] == "progress" then
			if r[1] == "text" then
				Rattus_Settings.progtext = not Rattus_Settings.progtext
				Debug(string.format("Showing progress as text: %s", tostring(Rattus_Settings.progtext)), RELEASE)
				if Rattus_Settings.progtext == true then
					RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go.png")
				else
					RX.UIElements.header.text:SetTexture(addon.identifier, "textures/font_go_bw.png")
				end
				rtx_ToggleComp = true
			elseif r[1] == "bar" then
				Rattus_Settings.progbar = not Rattus_Settings.progbar
				Debug(string.format("Showing progress bar: %s", tostring(Rattus_Settings.progbar)), RELEASE)
				if Rattus_Settings.progbar == true then
					RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar.png")
				else
					RX.UIElements.header.bars:SetTexture(addon.identifier, "textures/chart_bar_bw.png")
				end
				rtx_ToggleComp = true
			else
				if numargs >= 2 then
					Debug(string.format("Unknown mode for progress: %s", r[1]), RELEASE)
				else
					Debug(string.format("Need to specify text|bar for progress mode"), RELEASE)
				end
			end
		elseif r[0] == "options" then
			Rattus_Settings.mmoptions = not Rattus_Settings.mmoptions
			if Rattus_Settings.mmoptions then
				Debug("Options bar on minimap icon", RELEASE)
				RX.UIElements.controlbar.frame:SetPoint("TOPLEFT", RX.UIElements.minimap, "CENTER")
			else
				Debug("Options bar on Rattus header", RELEASE)
				RX.UIElements.controlbar.frame:SetPoint("TOPLEFT", RX.UIElements.header.icon, "TOPRIGHT", 0,0)
			end
		else
			RX.showUsage()
		end
	else
		RX.showUsage()
	end
end

function RX.Event_libZoneChange_Player(h, zn, zid)
	RX.HandleZoneChange(zn)
end

function RX.Event_Unit_Availability_Full(h,t)
	for k,v in pairs(t) do
		if v == "player" then
			Command.Event.Detach(Event.Unit.Availability.Full, nil, nil, nil, addon.identifier)
			rtx_AchvList = true
		end
	end
end

---------------------
-- Register Events
---------------------

Command.Event.Attach(Event.System.Update.Begin, RX.Event_System_Update_Begin, "Event.System.Update.Begin")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, function(h,a) if a == addon.identifier then svLoaded = true end end, "Event.Addon.SavedVariables.Load.End")
Command.Event.Attach(Event.Achievement.Complete, RX.Event_Achievement_Complete, "Event.Achievement.Complete")
Command.Event.Attach(Event.Achievement.Update, RX.Event_Achievement_Update, "Event.Achievement.Update")
Command.Event.Attach(Library.libZoneChange.Player, RX.Event_libZoneChange_Player, "Library.LibZoneChange.Player")
Command.Event.Attach(Command.Slash.Register("rattus"), RX.Command_Slash_Register, "Command.Slash.Register")
Command.Event.Attach(Event.Unit.Availability.Full, RX.Event_Unit_Availability_Full, "Event.Unit.Availability.Full")

print(string.format("v%s loaded.", addon.toc.Version))
